"""Container for constants used in the KIE package"""
MLFLOW_REGISTERED_MODEL_TAG = "registered_to"
MLFLOW_SYSTEM_PROMPT_FILE = 'system_prompt.txt'
MLFLOW_TRAILING_PROMPT_FILE = 'trailing_prompt.txt'
MLFLOW_RESPONSE_FORMAT_FILE = 'response_format.json'
